; (load "delcont.scm")

(defmacro (sandbox . expr)
  (let ((err (gensym)))
    `(guard
      (,err ((begin (write-string "caught exception '")
		    (display ,err)
		    (write-string "'\n")
		    #t)))
      ,@expr)))

; for mzscheme, replace the above definition of sandbox with
;	(define (sandbox . expr) #t)
; it won't catch exceptions, but then mzscheme doesn't produce
; any, except for the intentional syntax error at the end

(write-string "all answers are checked against mzscheme\n")

(write-string "examples from community.schemewiki.org\n")

(write-string "r  should be 6\t\t")
(sandbox
 (display (+ 1 (RESET (+ 2 3))))
 (newline))

(write-string "r  should be (1 2)\t")
(sandbox
 (display (cons 1 (RESET (cons 2 '()))))
 (newline))

(write-string "r  should be 4\t\t")
(sandbox
 (display (+ 1 (RESET (+ 2 (SHIFT k 3)))))
 (newline))

(write-string "r  should be (1 3)\t")
(sandbox
 (display (cons 1 (RESET (cons 2 (SHIFT k (cons 3 '()))))))
 (newline))

(write-string "r  should be 10\t\t")
(sandbox
 (display (+ 1 (RESET (+ 2 (SHIFT k (+ 3 (k 4)))))))
 (newline))

(write-string "r  should be 14\t\t")
(sandbox
 (display (+ 1 (RESET (+ 2 (SHIFT k (+ 3 (k 5) (k 1)))))))
 (newline))

(write-string "r  should be (1 3 2 4)\t")
(sandbox
 (display (cons 1 (RESET (cons 2 (SHIFT k (cons 3 (k (cons 4 '()))))))))
 (newline))

(write-string "r  should be (1 3 2 2 4)\t")
(sandbox
 (display (cons 1 (RESET (cons 2 (SHIFT k (cons 3 (k (k (cons 4 '())))))))))
 (newline))

(write-string "multiple SHIFTs\n")

(write-string "r  should be 2\t\t")
(sandbox
 (RESET (display (begin (SHIFT k1 (k1 1)) (SHIFT k2 (k2 2)))))
 (newline))

; haskeem fails this -- restart issue
(write-string "r  should be 12\t\t")
(sandbox
 (RESET (begin (display (SHIFT k1 (k1 1))) (display (SHIFT k2 (k2 2)))))
 (newline))

(write-string "r  should be 12\t\t")
(sandbox
 (RESET (display (SHIFT k1 (k1 1))) (display (SHIFT k2 (k2 2))))
 (newline))

(write-string "r0 should be 134234\t")
(sandbox
 (RESET0 (begin (SHIFT0 k1 (begin (k1 (display 1)) (k1 (display 2))))
		(SHIFT0 k2 (begin (k2 (display 3)) (k2 (display 4))))))
 (newline))

(write-string "r0 should be 134234\t")
(sandbox
 (RESET0 (begin (SHIFT0 k1 (k1 (display 1)) (k1 (display 2)))
		(SHIFT0 k2 (k2 (display 3)) (k2 (display 4)))))
 (newline))

(write-string "r0 should be 134234\t")
(sandbox
 (RESET0 (SHIFT0 k1 (k1 (display 1)) (k1 (display 2)))
	 (SHIFT0 k2 (k2 (display 3)) (k2 (display 4))))
 (newline))

(write-string "r0 should be 13564562356456\n\t     ")
(sandbox
 (RESET0 (begin (SHIFT0 k1 (begin (k1 (display 1)) (k1 (display 2))))
		(SHIFT0 k2 (begin (k2 (display 3)) (k2 (display 4))))
		(SHIFT0 k3 (begin (k3 (display 5)) (k3 (display 6))))))
 (write-string "\n\t     ")
 (RESET0 (begin (SHIFT0 k1 (k1 (display 1)) (k1 (display 2)))
		(SHIFT0 k2 (k2 (display 3)) (k2 (display 4)))
		(SHIFT0 k3 (k3 (display 5)) (k3 (display 6)))))
 (write-string "\n\t     ")
 (RESET0 (SHIFT0 k1 (k1 (display 1)) (k1 (display 2)))
	 (SHIFT0 k2 (k2 (display 3)) (k2 (display 4)))
	 (SHIFT0 k3 (k3 (display 5)) (k3 (display 6))))
 (newline))

(write-string "r  should be 57\t\t")
(sandbox
 (display (+ 1 (RESET (+ 2 (SHIFT k1 (+ 3
					(k1 5)
					(k1 1)
					(RESET (* 10 (SHIFT k2 (+ 3 (k2 4)))))
					))))))
 (newline))

(write-string "r  should be 59\t\t")
(sandbox
 (display (+ 1 (RESET (+ 2 (SHIFT k1 (+ 3
					(k1 5)
					(k1 1)
					(RESET (* 10 (SHIFT k2 (+ (k1 3)
								  (k2 4)))))
					))))))
 (newline))

; haskeem fails the following nine -- restart issue

(write-string "r  should be 1 3 2\t")
(sandbox
 (RESET (begin (write-string "1 ")
	       (SHIFT c (begin (c 'ignore)
			       (write-string "2 ")))
	       (write-string "3 ")))
 (newline))
(write-string "r  should be 1 3 2\t")
(sandbox
 (RESET (begin (write-string "1 ")
	       (SHIFT c (c 'ignore) (write-string "2 "))
	       (write-string "3 ")))
 (newline))
(write-string "r  should be 1 3 2\t")
(sandbox
 (RESET (write-string "1 ")
	(SHIFT c (c 'ignore) (write-string "2 "))
	(write-string "3 "))
 (newline))

(write-string "r  should be 1 2 3\t")
(sandbox
 (RESET (begin (write-string "1 ")
	       (SHIFT c (begin (write-string "2 ")
			       (c 'ignore)))
	       (write-string "3 ")))
 (newline))
(write-string "r  should be 1 2 3\t")
(sandbox
 (RESET (begin (write-string "1 ")
	       (SHIFT c (write-string "2 ") (c 'ignore))
	       (write-string "3 ")))
 (newline))
(write-string "r  should be 1 2 3\t")
(sandbox
 (RESET (write-string "1 ")
	(SHIFT c (write-string "2 ") (c 'ignore))
	(write-string "3 "))
 (newline))

(write-string "r  should be 1 3 2 3\t")
(sandbox
 (RESET (begin (write-string "1 ")
	       (SHIFT c (begin (c 'ignore)
			       (write-string "2 ")
			       (c 'ignore)))
	       (write-string "3 ")))
 (newline))
(write-string "r  should be 1 3 2 3\t")
(sandbox
 (RESET (begin (write-string "1 ")
	       (SHIFT c (c 'ignore) (write-string "2 ") (c 'ignore))
	       (write-string "3 ")))
 (newline))
(write-string "r  should be 1 3 2 3\t")
(sandbox
 (RESET (write-string "1 ")
	(SHIFT c (c 'ignore) (write-string "2 ") (c 'ignore))
	(write-string "3 "))
 (newline))

(write-string "from Shan, \"Shift to Control\"\n")

(write-string "r  should be (a)\t")
(sandbox
 (display (RESET (cons 'a (RESET (SHIFT f (SHIFT g '()))))))
 (newline))

(write-string "r0 should be ()\t\t")
(sandbox
 (display (RESET0 (cons 'a (RESET0 (SHIFT0 f (SHIFT0 g '()))))))
 (newline))

(write-string "r  should be (a)\t")
(sandbox
 (display (RESET (let ((y (SHIFT f (cons 'a (f '()))))) (SHIFT g y))))
 (newline))

(write-string "p  should be ()\t\t")
(sandbox
 (display (PROMPT (let ((y (CONTROL f (cons 'a (f '()))))) (CONTROL g y))))
 (newline))

; Shan sez this is an infinite loop... Kiselyov turns this into a printing
; loop by adding (display) somewhere, have to see if I can replicate that.
; I dunno if the name has to be 'f' in both CONTROL parts

;(PROMPT (begin (CONTROL f (begin (f 0) (f 0)))
;	       (CONTROL f (begin (f 0) (f 0)))))

;;;(write-string "should be 132342344234442344442344444234444442344...\n\t  ")
;;; with RESET2 it /is/ an infinite loop, just not the desired one...
;;; it produces 1313131313... instead
;;; I think this is actually the same partially-evaluated-thunk issue
;;; as with the other tests which are failing
;;;(PROMPT (begin (CONTROL f (begin (f (display 1)) (f (display 2))))
;;;	      (CONTROL f (begin (f (display 3)) (f (display 4))))))
;;;(newline)

(write-string "r  should be (a)\t")
(sandbox
 (display (RESET (let ((y (SHIFT f (cons 'a (f '()))))) (SHIFT g y))))
 (newline))

(write-string "r0 should be (a)\t")
(sandbox
 (display (RESET0 (let ((y (SHIFT0 f (cons 'a (f '()))))) (SHIFT0 g y))))
 (newline))

(write-string "p  should be ()\t\t")
(sandbox
 (display (PROMPT (let ((y (CONTROL f (cons 'a (f '()))))) (CONTROL g y))))
 (newline))

(write-string "p0 should be ()\t\t")
(sandbox
 (display
  (RESET (PROMPT0 (let ((y (CONTROL0 f (cons 'a (f '()))))) (CONTROL0 g y)))))
 (newline))

(write-string "r  should be (a b)\t")
(sandbox
 (display
  (RESET
   (RESET (cons 'a (RESET
		    (let ((y (SHIFT f (SHIFT g (cons 'b (f '()))))))
		      (SHIFT h y)))))))
 (newline))

(write-string "p  should be (a)\t")
(sandbox
 (display
  (PROMPT
   (PROMPT (cons 'a (PROMPT
		     (let ((y (CONTROL f (CONTROL g (cons 'b (f '()))))))
		       (CONTROL h y)))))))
 (newline))

(write-string "r0 should be (b)\t")
(sandbox
 (display
  (RESET0
   (RESET0 (cons 'a (RESET0
		     (let ((y (SHIFT0 f (SHIFT0 g (cons 'b (f '()))))))
		       (SHIFT0 h y)))))))
 (newline))

; haskeem fails this -- UNKNOWN ISSUE! TODO: investigate!

(write-string "p0 should be ()\t\t")
(sandbox
 (display
  (PROMPT0
   (PROMPT0 (cons 'a (PROMPT0
		      (let ((y (CONTROL0 f (CONTROL0 g (cons 'b (f '()))))))
			(CONTROL0 h y)))))))
 (newline))

(write-string "two-armed tests\n")
(write-string "test 1: non-trivial SHIFT only in false arm of if\n")
(define (doit1 flag)
  (if flag
      (write-string "r  should be (1 3)\t")
      (write-string "r  should be (1 4 2 5)\t"))
  (sandbox
   (display
    (cons 1 (RESET (cons 2 (if flag
			       (SHIFT k1 (cons 3 '()))
			       (SHIFT k2 (cons 4 (k2 (cons 5 '())))))))))
   (newline)))

(doit1 #f)
(doit1 #t)

(write-string "test 2: reverse of previous\n")
(define (doit2 flag)
  (if flag
      (write-string "r  should be (1 4 2 5)\t")
      (write-string "r  should be (1 3)\t"))
  (sandbox
   (display
    (cons 1 (RESET (cons 2 (if flag
			       (SHIFT k2 (cons 4 (k2 (cons 5 '()))))
			       (SHIFT k1 (cons 3 '())))))))
   (newline)))

(doit2 #f)
(doit2 #t)

(write-string "test 3: wrong? expansion of test 1\n")

(define (doit3 flag)
  (write-string "r  should be (1 4 3)\t")	; independent of flag
  (sandbox
   (display
    (cons 1
	  (let ((k2 (lambda (v2)
		      (let ((k1 (lambda (v1)
				  (RESET (cons 2 (if flag v1 v2))))))
			(cons 3 '())))))
	    (RESET (cons 4 (k2 (cons 5 '())))))))
   (newline)))

(doit3 #f)
(doit3 #t)

(write-string "test 4: expansion of test 2\n")
(define (doit4 flag)
  (if flag
      (write-string "r  should be (1 4 2 3)\t")
      (write-string "r  should be (1 4 2 5)\t"))
  (sandbox
   (display
    (cons 1
	  (let ((k2 (lambda (v2)
		      (RESET (cons 2 (if flag (cons 3 '()) v2))))))
	    (RESET (cons 4 (k2 (cons 5 '())))))))
   (newline)))

(doit4 #f)
(doit4 #t)

(write-string "test 5: modified version, with (SHIFT) in both arms of (if)\n")
(define (doit5 flag)
  (if flag
      (write-string "r  should be (1 3 2 4)\t")
      (write-string "r  should be (1 5 2 6)\t"))
  (sandbox
   (display
    (cons 1
	  (RESET (cons 2 (if flag
			     (SHIFT k1 (cons 3 (k1 (cons 4 '()))))
			     (SHIFT k2 (cons 5 (k2 (cons 6 '())))))))))
   (newline)))

(doit5 #f)
(doit5 #t)

(write-string
 "test 6: modified version of test 5, with (if) lifted out of (RESET)\n")

(define (doit6 flag)
  (if flag
      (write-string "r  should be (1 3 2 4)\t")
      (write-string "r  should be (1 5 2 6)\t"))
  (sandbox
   (display
    (cons 1
	  (if flag
	      (RESET (cons 2 (SHIFT k1 (cons 3 (k1 (cons 4 '()))))))
	      (RESET (cons 2 (SHIFT k2 (cons 5 (k2 (cons 6 '())))))))))
   (newline)))

(doit6 #f)
(doit6 #t)

(write-string "examples from Queinnec, augmented to test all versions\n")

(write-string "PROMPT/CONTROL examples\n")
(write-string "should be 3\t\t")
(sandbox
 (display (PROMPT (* 2 (CONTROL f 3))))
 (newline))

(write-string "should be 30\t\t")
(sandbox
 (display (PROMPT (* 2 (CONTROL f (* 5 (f 3))))))
 (newline))

(write-string "should be 12\t\t")
(sandbox
 (display (PROMPT (* 2 (CONTROL f (f (f 3))))))
 (newline))

(write-string "should be 6\t\t")
(sandbox
 (display ((PROMPT (* 2 (CONTROL f f))) 3))
 (newline))

(write-string "should be 35\t\t")
(sandbox
 (display (PROMPT (* 5
		     (PROMPT (* 2
				(CONTROL f2 (* 3 (CONTROL f3 7))))))))
 (newline))

; haskeem quasi-fails this: because it's a (lambda () ...), it escapes,
; and so at the end there's an uncaught exception which however carries
; along the correct value as its payload. So I think haskeem would handle
; this correctly if there were an implicit top-level catcher

(write-string "should be 7\t\t")
(sandbox
 (display (PROMPT (* 5 (
			(PROMPT (* 2
				   (CONTROL f2 (lambda ()
						 (* 3 (CONTROL f3 7))))))))))
 (newline))

(write-string "should be 21\t\t")
(sandbox
 (display (PROMPT (* 5
		     ((lambda (x) (CONTROL f1 x))
		      (* 3 (CONTROL f2 (* 2 (f2 7))))))))
 (newline))

(write-string "PROMPT0/CONTROL0 examples\n")
(write-string "should be 3\t\t")
(sandbox
 (display (PROMPT0 (* 2 (CONTROL0 f 3))))
 (newline))

(write-string "should be 30\t\t")
(sandbox
 (display (PROMPT0 (* 2 (CONTROL0 f (* 5 (f 3))))))
 (newline))

(write-string "should be 12\t\t")
(sandbox
 (display (PROMPT0 (* 2 (CONTROL0 f (f (f 3))))))
 (newline))

(write-string "should be 6\t\t")
(sandbox
 (display ((PROMPT0 (* 2 (CONTROL0 f f))) 3))
 (newline))

(write-string "should be 7\t\t")
(sandbox
 (display (PROMPT0 (* 5
		      (PROMPT0 (* 2
				  (CONTROL0 f2 (* 3 (CONTROL0 f3 7))))))))
 (newline))

(write-string "should be 7\t\t")
(sandbox
 (display (PROMPT0 (* 5 (
			(PROMPT0 (* 2
				   (CONTROL0 f2 (lambda ()
						 (* 3 (CONTROL0 f3 7))))))))))
 (newline))

; There's something odd with this test: in mzscheme, it prints the
; result 21, then the whole rest of the test vanishes; apparently
; enough of the surrounding context is sucked up that the rest of the
; world just vanishes. If I surround this with a (RESET ...), the
; test continues, but this doesn't print anything. Hmmm!

(write-string "should be 21\t\t")
(sandbox
 (display (PROMPT0 (* 5
		      ((lambda (x) (CONTROL0 f1 x))
		       (* 3 (CONTROL0 f2 (* 2 (f2 7))))))))
 (newline))

(write-string "RESET/SHIFT examples\n")
(write-string "should be 3\t\t")
(sandbox
 (display (RESET (* 2 (SHIFT f 3))))
 (newline))

(write-string "should be 30\t\t")
(sandbox
 (display (RESET (* 2 (SHIFT f (* 5 (f 3))))))
 (newline))

(write-string "should be 12\t\t")
(sandbox
 (display (RESET (* 2 (SHIFT f (f (f 3))))))
 (newline))

(write-string "should be 6\t\t")
(sandbox
 (display ((RESET (* 2 (SHIFT f f))) 3))
 (newline))

(write-string "should be 35\t\t")
(sandbox
 (display (RESET (* 5
		    (RESET (* 2
			      (SHIFT f2 (* 3 (SHIFT f3 7))))))))
 (newline))

(write-string "should be 7\t\t")
(sandbox
 (display (RESET (* 5 (
		       (RESET (* 2
				 (SHIFT f2 (lambda ()
					     (* 3 (SHIFT f3 7))))))))))
 (newline))

(write-string "should be 42\t\t")
(sandbox
 (display (RESET (* 5
		    ((lambda (x) (SHIFT f1 x))
		     (* 3 (SHIFT f2 (* 2 (f2 7))))))))
 (newline))

(write-string "RESET0/SHIFT0 examples\n")
(write-string "should be 3\t\t")
(sandbox
 (display (RESET0 (* 2 (SHIFT0 f 3))))
 (newline))

(write-string "should be 30\t\t")
(sandbox
 (display (RESET0 (* 2 (SHIFT0 f (* 5 (f 3))))))
 (newline))

(write-string "should be 12\t\t")
(sandbox
 (display (RESET0 (* 2 (SHIFT0 f (f (f 3))))))
 (newline))

(write-string "should be 6\t\t")
(sandbox
 (display ((RESET0 (* 2 (SHIFT0 f f))) 3))
 (newline))

(write-string "should be 7\t\t")
(sandbox
 (display (RESET0 (* 5
		     (RESET0 (* 2
				(SHIFT0 f2 (* 3 (SHIFT0 f3 7))))))))
 (newline))

(write-string "should be 7\t\t")
(sandbox
 (display (RESET0 (* 5 (
		       (RESET0 (* 2
				 (SHIFT0 f2 (lambda ()
					     (* 3 (SHIFT0 f3 7))))))))))
 (newline))

(write-string "should be 42\t\t")
(sandbox
 (display (RESET0 (* 5
		     ((lambda (x) (SHIFT0 f1 x))
		      (* 3 (SHIFT0 f2 (* 2 (f2 7))))))))
 (newline))

(write-string "r  something actually useful! a small table of factorials\n")

; Eventually we want to be able to use code like this:

; (defmacro (range from step to)
;   (let ((f (gensym))
;	(i (gensym)))
;     `(SHIFT ,f
;	    (do ((,i ,from (+ ,i ,step)))
;		((> ,i ,to) #f)
;	      (,f ,i)))))
;
; (RESET (begin (display (my-fact (range 1 3 16))) (newline)))

; Just to show I've got nothing up my sleeve, here is a perfectly cromulent
; function; we could equally well use the built-in (factorial) function

(define (my-fact n)
  (if (<= n 0)
      1
      (* n (my-fact (- n 1)))))

(define from 1)
(define to 21)
(define step 3)

(sandbox
 (RESET (display (my-fact 
		  (SHIFT f
			 (do ((i from (+ i step)))
			     ((> i to) #f)
			   (write-string (number->string i))
			   (write-string "! =\t")
			   (f i)))))
	(newline)))

; This one is specific to haskeem, because the macro RESET doesn't start
; the thunk in the middle, but rather at the beginning; that makes the
; ping-pong possible

(write-string
 "These rely on the RESET thunk being re-evaluated from the beginning,\n")
(write-string
 "so strictly speaking both are wrong. So I don't know what the answers\n")
(write-string
 "\"should be\"... but they look plausible\n")
(define (doit7 flag)
  (sandbox
   (display
    (cons 1
	  (RESET (cons 2 (if flag
			     (SHIFT k1 (set! flag (not flag))
				    (cons 3 (k1 (cons 4 '()))))
			     (SHIFT k2 (set! flag (not flag))
				    (cons 5 (k2 (cons 6 '())))))))))
   (newline)))

; it's hard to say if these are right... both rely on a wrong implementation
; of the thunk stuff

(doit7 #f)	; (1 5 3 2 6) -- mzscheme sez (1 5 2 6)
(doit7 #t)	; (1 3 5 2 4) -- mzscheme sez (1 3 2 4)

(write-string
 "syntax error coming up! should be an error... this'll kill mzscheme\n")

(sandbox (display (+ 1 (RESET (+ 2 (SHIFT (+ k 1) 3))))))
(write-string
 "haskeem should have caught the exception raised by the syntax error\n")
